searchData={"content_type":"text/markdown","items":[{"doc":"Convenience macros for doing comparisons","ref":"CompareChain.html","title":"CompareChain","type":"module"},{"doc":"Valid expressions for `compare?/1` and `compare?/2` follow these three rules:","ref":"CompareChain.html#module-valid-expressions","title":"Valid expressions - CompareChain","type":"module"},{"doc":"At least one of these must be included: `<`, `>`, `<=`, `>=`, `==`, `!=`,\n`===`, or `!===`. So this is valid:\n\n    compare?(1 < 2 < 3)\n\nbut this is not:\n\n    compare?(true)\n\nbecause it does not contain any comparisons.","ref":"CompareChain.html#module-1-a-comparison-operator-like-must-be-present","title":"1. A comparison operator like `<` must be present. - CompareChain","type":"module"},{"doc":"The boolean operators `and`, `or`, and `not` are allowed in expressions so\nlong as all of their arguments (eventually) contain a comparison. So this is\nvalid:\n\n    compare?(1 < 2 < 3 and 4 < 5)\n\nas is this:\n\n    compare?(not (not 1 < 2 < 3))\n\nbut this is not:\n\n    compare?(1 < 2 < 3 and true)\n\nbecause the right side of `and` fails to contain a comparison. This\nexpression can be refactored to be valid by moving the non-comparison branch\noutside `compare?/1` like so:\n\n    compare?(1 < 2 < 3) and true","ref":"CompareChain.html#module-2-all-arguments-to-boolean-operators-must-also-be-valid-expressions","title":"2. All arguments to boolean operators must also be valid expressions. - CompareChain","type":"module"},{"doc":"So this is not valid:\n\n    compare?(my_function(a < b), Date)\n\nbecause its root operator is `my_function/1`. This expression can be\nrefactored to be valid by moving `compare?/2` inside `my_function/1` like so:\n\n    my_function(compare?(a < b, Date))\n\nWe restrict expressions in this fashion so we can guarantee that `compare?/1`\nand `compare?/2` will always return a boolean.\n\nAlso note that arguments to _comparisons_ may be arbitrarily complicated:\n\n    compare?(a < Date.utc_today(), Date)","ref":"CompareChain.html#module-3-the-root-operator-of-an-expression-must-be-a-comparison-or-a-boolean","title":"3. The root operator of an expression must be a comparison or a boolean. - CompareChain","type":"module"},{"doc":"Macro that performs chained comparison with operators like `<`.\n\nYou may also include the boolean operators `and`, `or`, and `not` in the\nexpression so long as all their arguments all (eventually) contain\ncomparisons. See the moduledoc for more details.\n\nFor a version that also does semantic comparison, see: `compare?/2`.","ref":"CompareChain.html#compare?/1","title":"CompareChain.compare?/1","type":"macro"},{"doc":"Chained comparison:\n\n    iex> import CompareChain\n    iex> compare?(1 < 2 < 3)\n    true\n\nComparisons joined by logical operators:\n\n    iex> import CompareChain\n    iex> compare?(1 >= 2 >= 3 or 4 >= 5 >= 6)\n    false","ref":"CompareChain.html#compare?/1-examples","title":"Examples - CompareChain.compare?/1","type":"macro"},{"doc":">","ref":"CompareChain.html#compare?/1-warnings-and-errors","title":"Warnings and errors - CompareChain.compare?/1","type":"macro"},{"doc":">\n> Expressions which compare matching structs like:\n>\n>     iex> compare?(~D[2017-03-31] < ~D[2017-04-01])\n>     false\n>\n> Will result in a warning:\n>\n> ```plain\n> ... [warning] Performing structural comparison on matching structs.\n>\n> Did you mean to use `compare?/2`?\n>\n>   compare?(~D[2017-03-31] < ~D[2017-04-01], Date)\n> ```\n>\n> You probably want to use `compare?/2`, which does semantic comparison,\n> instead.\n\n>","ref":"CompareChain.html#compare?/1-comparing-structs-will-warn-warning","title":"Comparing structs will warn {: .warning} - CompareChain.compare?/1","type":"macro"},{"doc":">\n> See the section on valid expressions in the moduledoc for details.","ref":"CompareChain.html#compare?/1-invalid-expressions-will-raise-error","title":"Invalid expressions will raise {: .error} - CompareChain.compare?/1","type":"macro"},{"doc":"Macro that performs chained, semantic comparison with operators like `<` by\nrewriting the expression using the `compare/2` function defined by the\nprovided module.\n\nThis is like how you can provide a module as the second argument to\n`Enum.sort/2` when you need to sort items semantically.\n\nYou may also include the boolean operators `and`, `or`, and `not` in the\nexpression so long as all their arguments all (eventually) contain\ncomparisons. See the moduledoc for more details.\n\nFor a version that does chained comparison using the normal `<` operators,\nsee: `compare?/1`.","ref":"CompareChain.html#compare?/2","title":"CompareChain.compare?/2","type":"macro"},{"doc":"Semantic comparison (note how `a   import CompareChain\n    iex> a = ~D[2017-03-31]\n    iex> b = ~D[2017-04-01]\n    iex> a   compare?(a   import CompareChain\n    iex> a = ~D[2017-03-31]\n    iex> b = ~D[2017-04-01]\n    iex> c = ~D[2017-04-02]\n    iex> compare?(a   import CompareChain\n    iex> a = ~T[15:00:00]\n    iex> b = ~T[16:00:00]\n    iex> c = ~T[17:00:00]\n    iex> compare?(a   c, Time)\n    false\n\nMore complex expressions:\n\n    iex> import CompareChain\n    iex> compare?(%{a: ~T[16:00:00]}.a <= ~T[17:00:00], Time)\n    true\n\nCustom module:\n\n    iex> import CompareChain\n    iex> defmodule AlwaysGreaterThan do\n    iex>   def compare(_left, _right), do: :gt\n    iex> end\n    iex> compare?(1 > 2 > 3, AlwaysGreaterThan)\n    true","ref":"CompareChain.html#compare?/2-examples","title":"Examples - CompareChain.compare?/2","type":"macro"},{"doc":">","ref":"CompareChain.html#compare?/2-warnings-and-errors","title":"Warnings and errors - CompareChain.compare?/2","type":"macro"},{"doc":">\n> Expressions which include either `===` or `!==` like:\n>\n>     iex> compare?(~D[2017-03-31] !== ~D[2017-04-01], Date)\n>     true\n>\n> Will result in a warning:\n>\n> ```plain\n> ... [warning] Performing semantic comparison using either: `===` or `!===`.\n> This is reinterpreted as `==` or `!=`, respectively.\n> ```\n>\n> These operators have no additional meaning over `==` and `!=` when doing\n> semantic comparison.\n\n>","ref":"CompareChain.html#compare?/2-using-the-strict-operators-will-warn-warning","title":"Using the \"strict\" operators will warn {: .warning} - CompareChain.compare?/2","type":"macro"},{"doc":">\n> See the section on valid expressions in the moduledoc for details.","ref":"CompareChain.html#compare?/2-invalid-expressions-will-raise-error","title":"Invalid expressions will raise {: .error} - CompareChain.compare?/2","type":"macro"},{"doc":"# CompareChain","ref":"readme.html","title":"CompareChain","type":"extras"},{"doc":"Provides convenience macros for comparisons which do:\n\n  * chained comparisons like: `a  `, `<=`, `>=`, `==`, `!=`, `===`, and `!==`\n  * combinations using: `and`, `or`, and `not`","ref":"readme.html#description","title":"Description - CompareChain","type":"extras"},{"doc":"```elixir\niex> import CompareChain\n\n# Chained comparisons\niex> compare?(1 < 2 < 3)\ntrue\n\n# Semantic comparisons\niex> compare?(~D[2017-03-31] < ~D[2017-04-01], Date)\ntrue\n\n# Chained, semantic comparisons\niex> compare?(~D[2017-03-31] < ~D[2017-04-01] < ~D[2017-04-02], Date)\ntrue\n\n# Semantic comparisons with logical operators\niex> compare?(~T[16:00:00] <= ~T[16:00:00] and not (~T[17:00:00] <= ~T[17:00:00]), Time)\nfalse\n\n# More complex expressions\niex> compare?(%{a: ~T[16:00:00]}.a <= ~T[17:00:00], Time)\ntrue\n```","ref":"readme.html#examples","title":"Examples - CompareChain","type":"extras"},{"doc":"Add `compare_chain` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:compare_chain, \"~> 0.5\"}\n  ]\nend\n```\n\nDocumentation can be found at  .","ref":"readme.html#installation","title":"Installation - CompareChain","type":"extras"},{"doc":"Once installed, you can add:\n\n```elixir\nimport CompareChain\n```\n\nto your `defmodule` and you will have access to `CompareChain.compare?/1` and `CompareChain.compare?/2`.","ref":"readme.html#usage","title":"Usage - CompareChain","type":"extras"},{"doc":"`CompareChain` was originally motivated by the following situation:\n\n> You have an interval of time bounded by a two `%Date{}` structs: `start_date` and `end_date`.\n> You want to know if some third `date` falls in that interval.\n> How do you write this?\n\nIn Elixir, we'd write this as follows:\n\n```elixir\nDate.compare(start_date, date) == :lt and\n  Date.compare(date, end_date) == :lt\n```\n\nThis is verbose and therefore a little hard to read.\nIt's also potentially incorrect, though not obviously so.\nWhat if `date` is considered \"within\" the interval even if it equals `start_date` or `end_date`?\nTo include the bounds in our comparison, we'd instead write the expression like this:\n\n```elixir\nDate.compare(start_date, date) != :gt and\n  Date.compare(date, end_date) != :gt\n```\n\n(We could have written `Date.compare(start_date, date) in [:lt, :eq]`, but `!= :gt` is faster.)\n\nIn order to spot the difference between these two cases, you have to keep several things in mind:\n\n  * The order of the arguments passed to `Date.compare/2`\n  * The specific comparison operators for each clause: `==` vs. `!=`\n  * The specific comparison atoms for each clause: `:lt` vs. `:gt`\n\nSince this is hard to read, it's easy to introduce bugs.\nContrast this with how you'd write the equivalent code in Python:\n\n```\nstart_date <  date <  end_date # excluding bounds\nstart_date <= date <= end_date # including bounds\n```\n\nThis is much easier to read.\nSo why can't we write this in Elixir?\nTwo reasons:\n\n  * Structural comparison operators\n  * Chained vs. nested comparisons","ref":"readme.html#background-and-motivation","title":"Background and motivation - CompareChain","type":"extras"},{"doc":"Operators like `<` do _structural_ comparison instead of _semantic_ comparison.\nFrom the [`Kernel` docs](https://hexdocs.pm/elixir/Kernel.html#module-structural-comparison):\n\n> ... **comparisons in Elixir are structural**, as it has the goal\n  of comparing data types as efficiently as possible to create flexible\n  and performant data structures. This distinction is specially important\n  for functions that provide ordering, such as `>/2`, `</2`, `>=/2`,\n  `<=/2`, `min/2`, and `max/2`. For example:\n>\n> ```elixir\n> ~D[2017-03-31] > ~D[2017-04-01]\n> ```\n>\n> will return `true` because structural comparison compares the `:day`\n  field before `:month` or `:year`. In order to perform semantic comparisons,\n  the relevant data-types provide a `compare/2` function, such as\n  `Date.compare/2`:\n>\n> ```elixir\n> iex> Date.compare(~D[2017-03-31], ~D[2017-04-01])\n> :lt\n> ```\n\nIn other words, although `~D[2017-03-31] > ~D[2017-04-01]` is perfectly valid code, it does _not_ tell you if `~D[2017-03-31]` is a later date than `~D[2017-04-01]` like you might expect.\nInstead, you need to use `Date.compare/2`.","ref":"readme.html#structural-comparison-operators","title":"Structural comparison operators - CompareChain","type":"extras"},{"doc":"Additionally, even if `~D[2017-03-31] > ~D[2017-04-01]` did do semantic comparison, you still couldn't write the interval check like you do in Python.\nThis is because in Python, an expression like `1 < 2 < 3` is syntactic sugar for `(1 < 2) and (2 < 3)`, aka a series of \"chained\" expressions.\n\nElixir does not provide an equivalent syntactic sugar.\nInstead, `1 < 2 < 3` is evaluated as `(1 < 2) < 3`, aka a series of \"nested\" expressions.\nSince `(1 < 2) < 3` simplifies to `true < 3`, that's probably not what you want!\n\nElixir will even warn you when you attempt an expression like that:\n\n> warning: Elixir does not support nested comparisons. Something like\n>\n>      x  \n> is equivalent to\n>\n>      (x  \n> which ultimately compares z with the boolean result of (x  \n>      x < y and y < z","ref":"readme.html#chained-vs-nested-comparisons","title":"Chained vs. nested comparisons - CompareChain","type":"extras"},{"doc":"`CompareChain` attempts to address both of these issues with the macro `CompareChain.compare?/2`.\nIts job is to take code similar to how you'd like to write it and rewriting it to be semantically correct.\n\nFor our motivating example, we'd write this:\n\n```elixir\nimport CompareChain\ncompare?(start_date <  date <  end_date, Date) # excluding bounds\ncompare?(start_date <= date <= end_date, Date) # including bounds\n```\n\nAnd at compile time, `CompareChain.compare?/2` rewrites those to be:\n\n```elixir\n# excluding bounds\nDate.compare(start_date, date) == :lt and\n  Date.compare(date, end_date) == :lt\n\n# including bounds\nDate.compare(start_date, date) != :gt and\n  Date.compare(date, end_date) != :gt\n```\n\nThis way your code is more readable while still remaining correct.\n\n`CompareChain.compare?/1` is also available in case you only need chained comparison using the structural operators:\n\n```elixir\ncompare?(1 < 2 < 3)\n```\n\nThough I find this case comes up less often.","ref":"readme.html#comparechain","title":"CompareChain - CompareChain","type":"extras"},{"doc":"As a happy accident, `CompareChain.compare?/2` always uses fewer characters than its `compare/2` counterpart:\n\n```elixir\ncompare?(a <= b, Date)\n# vs.\nDate.compare(a, b) != :gt\n```\n\n(Assuming you've already included `import CompareChain`, of course!)\n\nBecause it's shorter _and_ more readable, these days I always use `CompareChain` for any semantic comparison, chained or not.","ref":"readme.html#one-last-selling-point","title":"One last selling point - CompareChain","type":"extras"},{"doc":"# CHANGELOG","ref":"changelog.html","title":"CHANGELOG","type":"extras"},{"doc":"* Allow `===` and `!===` in expressions\n  * Improve documentation\n  * Fix bug with certain operation chains:\n    ```\n    compare?(1 < 2 != 3 < 4) #=> true\n    compare?(1 < 2 != 3 > 4) #=> true (wrong!)\n    ```\n  * Improve error message and documentation for invalid expressions\n  * [BREAKING] all branches of `and`, `or` and `not` must now contain comparisons\n    * Example: `compare?(1 < 2 and true)` used to be ok but is no longer\n      allowed because the right argument to `and` doesn't contain a comparison.","ref":"changelog.html#v0-5-0-2024-05-22","title":"v0.5.0 (2024-05-22) - CHANGELOG","type":"extras"},{"doc":"* Warn when `compare?/1` is used on a struct","ref":"changelog.html#v0-4-0-2023-09-10","title":"v0.4.0 (2023-09-10) - CHANGELOG","type":"extras"},{"doc":"* Allow `==` and `!=` in expressions\n  * Allow `Elixir >= 1.13.0`\n    * `Macro.prewalker` was introduced in `1.13.0`","ref":"changelog.html#v0-3-0-2023-01-28","title":"v0.3.0 (2023-01-28) - CHANGELOG","type":"extras"},{"doc":"* Allow `not` in expressions","ref":"changelog.html#v0-2-0-2022-11-05","title":"v0.2.0 (2022-11-05) - CHANGELOG","type":"extras"},{"doc":"* Initial release","ref":"changelog.html#v0-1-0-2022-11-03","title":"v0.1.0 (2022-11-03) - CHANGELOG","type":"extras"}],"producer":{"name":"ex_doc","version":[48,46,51,50,46,50]}}